from jinja2 import Markup

from . import common


def sorting_anchor(request, field_name, title):
    return Markup(common.render_sort_anchor(request, field_name, title))


def sort_queryset(request, queryset):
    order_by = common.get_order_by_from_request(request)
    return common.sort_queryset(queryset, order_by)



from django.conf import settings

DEFAULT_SORT_UP = getattr(settings, "DEFAULT_SORT_UP", " &uarr;")
DEFAULT_SORT_DOWN = getattr(settings, "DEFAULT_SORT_DOWN", " &darr;")
INVALID_FIELD_RAISES_404 = getattr(settings, "SORTING_INVALID_FIELD_RAISES_404", False)

SORT_DIRECTIONS = {
    "asc": {"icon": DEFAULT_SORT_UP, "next": "desc"},
    "desc": {"icon": DEFAULT_SORT_DOWN, "next": ""},
    "": {"icon": "", "next": "asc"},
}



from django import template
from django.http import Http404
from django.utils.translation import gettext_lazy as _

from .. import common
from ..settings import INVALID_FIELD_RAISES_404

register = template.Library()


def anchor(parser, token):
    """
    Parses a tag that's supposed to be in this format '{% anchor field title %}'
    Title may be a "string", _("trans string"), or variable
    """
    bits = [b for b in token.split_contents()]
    if len(bits) < 2:
        raise template.TemplateSyntaxError("anchor tag takes at least 1 argument.")

    title_is_var = False
    title_is_translatable = False
    try:
        title = bits[2]
        if title[0] in ('"', "'"):
            if title[0] == title[-1]:
                title = title[1:-1]
            else:
                raise template.TemplateSyntaxError(
                    'anchor tag title must be a "string", _("trans string"), or variable'
                )
        elif title.startswith('_("') or title.startswith("_('"):
            title_is_translatable = True
        else:
            title_is_var = True
    except IndexError:
        title = bits[1].capitalize()

    return SortAnchorNode(
        bits[1].strip(), title.strip(), title_is_var, title_is_translatable
    )


class SortAnchorNode(template.Node):
    """
    Renders an <a> HTML tag with a link which href attribute
    includes the field on which we sort and the direction.
    and adds an up or down arrow if the field is the one
    currently being sorted on.

    Eg.
        {% anchor name Name %} generates
        <a href="/the/current/path/?sort=name&dir=asc" title="Name">Name</a>

    """

    def __init__(self, field, title, title_is_var, title_is_translatable):
        self.field = field
        self.title = title
        self.title_is_var = title_is_var
        self.title_is_translatable = title_is_translatable

    def render(self, context):
        if self.title_is_var:
            display_title = context[self.title]
        elif self.title_is_translatable:
            display_title = _(self.title[3:-2])
        else:
            display_title = self.title

        return common.render_sort_anchor(context["request"], self.field, display_title)


def autosort(parser, token):
    bits = [b.strip("\"'") for b in token.split_contents()]
    help_msg = (
        "autosort tag synopsis: {%% autosort queryset [as " "context_variable] %%}"
    )
    context_var = None

    # Check if has not required "as new_context_var" part
    if len(bits) == 4 and bits[2] == "as":
        context_var = bits[3]
        del bits[2:]

    if len(bits) != 2:
        raise template.TemplateSyntaxError(help_msg)

    return SortedDataNode(bits[1], context_var=context_var)


class SortedDataNode(template.Node):
    """
    Automatically sort a queryset with {% autosort queryset %}
    """

    def __init__(self, queryset_var, context_var=None):
        self.queryset_var = template.Variable(queryset_var)
        self.context_var = context_var

    def render(self, context):
        if self.context_var is not None:
            key = self.context_var
        else:
            key = self.queryset_var.var

        queryset = self.queryset_var.resolve(context)
        order_by = common.get_order_by_from_request(context["request"])

        try:
            context[key] = common.sort_queryset(queryset, order_by)
        except ValueError as e:
            raise template.TemplateSyntaxError from e
        except AttributeError:
            if INVALID_FIELD_RAISES_404:
                raise Http404(
                    "Invalid field sorting. If INVALID_FIELD_RAISES_404 were set to "
                    "False, the error would have been ignored."
                )
            context[key] = queryset

        return ""


anchor = register.tag(anchor)
autosort = register.tag(autosort)



"""
Common to Django tags (sorting_tags) and Jinja2 globals (jinja2_globals)
"""
from operator import attrgetter

from .settings import SORT_DIRECTIONS


def render_sort_anchor(request, field_name, title):
    get_params = request.GET.copy()
    sort_by = get_params.get("sort", None)
    if sort_by == field_name:
        # Render anchor link to next direction
        current_direction = SORT_DIRECTIONS[get_params.get("dir", "")]
        icon = current_direction["icon"]
        next_direction_code = current_direction["next"]
    else:
        icon = ""
        next_direction_code = "asc"

    # Not usual dict (can't update to replace)
    get_params["sort"] = field_name
    get_params["dir"] = next_direction_code
    url_append = "?" + get_params.urlencode() if get_params else ""
    return f'<a href="{request.path}{url_append}" title="{title}">{title}{icon}</a>'


def get_order_by_from_request(request):
    """
    Retrieve field used for sorting a queryset

    :param request: HTTP request
    :return: the sorted field name, prefixed with "-" if ordering is descending
    """
    sort_direction = request.GET.get("dir")
    field_name = (request.GET.get("sort") or "") if sort_direction else ""
    sort_sign = "-" if sort_direction == "desc" else ""
    return f"{sort_sign}{field_name}"


def need_python_sorting(queryset, order_by):
    if order_by.find("__") >= 0:
        # Python can't sort order_by with '__'
        return False

    # Python sorting if not a DB field
    field = order_by[1:] if order_by[0] == "-" else order_by
    field_names = [f.name for f in queryset.model._meta.get_fields()]
    return field not in field_names


def sort_queryset(queryset, order_by):
    """order_by is an Django ORM order_by argument"""
    if not order_by:
        return queryset

    if need_python_sorting(queryset, order_by):
        # Fallback on pure Python sorting (much slower on large data)

        # The field name can be prefixed by the minus sign and we need to
        # extract this information if we want to sort on simple object
        # attributes (non-model fields)
        if order_by[0] == "-":
            if len(order_by) == 1:
                # Prefix without field name
                raise ValueError

            reverse = True
            name = order_by[1:]
        else:
            reverse = False
            name = order_by
        if hasattr(queryset[0], name):
            return sorted(queryset, key=attrgetter(name), reverse=reverse)
        else:
            raise AttributeError
    else:
        return queryset.order_by(order_by)
